## rich-click compat for rich-less support
from typing import Any, Type

__all__ = ["click", "Console", "ClickGroup", "Table"]

try:
    import rich_click as click  # type: ignore
    from rich.console import Console  # type: ignore
    from rich.table import Table  # type: ignore

    from rich_click.rich_group import RichGroup  # type: ignore

    ClickGroup = RichGroup  # type: ignore[no-redef]
except ImportError:
    import click  # type: ignore[no-redef]
    from click import Group

    ClickGroup: Type[Group] = Group  # type: ignore[no-redef]

    class Console:  # type: ignore[no-redef]
        def __init__(self) -> None:
            pass

        def print(self, text: Any, **kwargs) -> None:
            if isinstance(text, Table):
                click.echo(text)
            else:
                print(text)

    class Table:  # type: ignore[no-redef]
        __slots__ = ("cols", "rows", "styles")

        def __init__(self) -> None:
            self.cols: list[str] = []
            self.rows: list[list[str]] = []
            self.styles: list[str] = []

        def add_column(self, col: str, style: str = "") -> None:
            self.cols.append(col)
            self.styles.append(style)

        def add_row(self, *row) -> None:
            self.rows.append([*row])

        def __str__(self) -> str:
            text = "".join([click.style(f"{col:10}", fg=self.styles.pop(0)) for col in self.cols]) + "\n"
            for row in self.rows:
                for col in row:
                    text += f"{col:10}"
                text += "\n"
            return text


def print_rich_table(cols: list[str], rows: list[list[str]], styles: list[str]) -> None:
    table = Table()
    for style, col in zip(styles, cols):
        table.add_column(col, style=style)
    for row in rows:
        table.add_row(*row)

    Console().print(table)



from .transformer import Transformer
from .loader import Loader
from .homer import save_metadata
from .helper import out
from .builder import run_cli as cli_runner

import time
from datetime import datetime, timedelta
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler, FileModifiedEvent
from pathlib import Path
import threading


class CLIManifestReloader(FileSystemEventHandler):
    def __init__(self, manifest_path: str, run_cli: bool, run_cli_args: tuple[str]) -> None:
        self.manifest_path = manifest_path
        self.run_cli = run_cli
        self.run_cli_args = run_cli_args
        self.last_modified = datetime.now()

        super().__init__()

    def on_modified(self, event):
        if event.is_directory or not event.src_path.endswith(self.manifest_path):
            return

        if not isinstance(event, FileModifiedEvent):
            return

        if datetime.now() - self.last_modified < timedelta(seconds=1):
            return

        self.last_modified = datetime.now()

        t = threading.Thread(target=self.reload, args=(self.manifest_path, self.run_cli, self.run_cli_args))
        t.daemon = True
        t.start()

    @classmethod
    def watch(cls, manifest_path: str, run_cli: bool, run_cli_args: tuple[str]):
        event_handler = cls(manifest_path, run_cli, run_cli_args)
        observer = Observer()
        observer.schedule(event_handler, path=Path(manifest_path).parent, recursive=False)
        observer.start()

        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            observer.stop()
        observer.join()

    @staticmethod
    def reload(manifest_path: str, run_cli: bool, run_cli_args: tuple[str]):
        manifest_io = open(manifest_path, "r")

        T = Transformer(manifest_io)
        Loader.load_from_cli(T.cli)
        save_metadata(manifest_io.name, T.cli)
        out(f"✨ Reloaded {T.cli.name} CLI v{T.cli.version} ✨", fg="green")

        if run_cli:
            cli_runner(T.cli.name, T.cli.code, run_cli_args)



## CLI to generate CLIs
import contextlib
from io import TextIOWrapper
import os
from typing import IO, Any, Optional, TextIO, Union, cast

from click.core import Context, Parameter
from click.types import _is_file_like

from .rich import click, Console, print_rich_table
from .rich import ClickGroup  # type: ignore

from .builder import build_cli, build_cli_from_manifest, run_cli
from .helper import (
    CLIFFY_CLI_DIR,
    age_datetime,
    exit_err,
    indent_block,
    out,
    out_err,
    write_to_file,
)
from .homer import get_clis, get_metadata, get_metadata_path, remove_metadata, save_metadata
from .loader import Loader
from .manifests import Manifest, set_manifest_version
from .transformer import Transformer
from .reloader import CLIManifestReloader

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


class AliasedGroup(ClickGroup):
    def get_command(self, ctx: click.Context, cmd_name: Optional[str]) -> Optional[click.Command]:
        with contextlib.suppress(KeyError):
            cmd_name = ALIASES[cmd_name].name  # type: ignore
        return super().get_command(ctx, cmd_name or "")


class ManifestOrCLI(click.File):
    def convert(  # type: ignore[override]
        self, value: Union[str, "os.PathLike[str]", IO[Any]], param: Optional[Parameter], ctx: Optional[Context]
    ) -> Union[str, IO[Any]]:
        if _is_file_like(value):
            return value

        value = cast("Union[str, os.PathLike[str]]", value)

        if isinstance(value, os.PathLike) or value.endswith("yaml"):
            return super().convert(value, param, ctx)

        return value


@click.group(context_settings=CONTEXT_SETTINGS, cls=AliasedGroup)  # type: ignore[arg-type]
@click.version_option()
def cli() -> None:
    pass


@cli.command()  # type: ignore[arg-type]
@click.argument("manifests", type=click.File("rb"), nargs=-1)
def load(manifests: list[TextIO]) -> None:
    """Load CLI for given manifest(s)"""
    for manifest in manifests:
        T = Transformer(manifest)
        Loader.load_from_cli(T.cli)
        save_metadata(manifest.name, T.cli)
        out(f"✨ Generated {T.cli.name} CLI v{T.cli.version} ✨", fg="green")
        out("$", fg="magenta", nl=False)
        out(f" {T.cli.name} -h")


@cli.command()  # type: ignore[arg-type]
@click.argument("cli_names", type=str, nargs=-1)
def update(cli_names: list[str]) -> None:
    """Reloads CLI by name"""
    for cli_name in cli_names:
        if cli_metadata := get_metadata(cli_name):
            T = Transformer(open(cli_metadata.runner_path, "r"))
            Loader.load_from_cli(T.cli)
            save_metadata(cli_metadata.runner_path, T.cli)
            out(f"✨ Reloaded {T.cli.name} CLI v{T.cli.version} ✨", fg="green")
            out("$", fg="magenta", nl=False)
            out(f" {T.cli.name} -h")
        else:
            out_err(f"~ {cli_name} not found")


@cli.command()  # type: ignore[arg-type]
@click.argument("manifest", type=click.File("rb"))
def render(manifest: TextIO) -> None:
    """Render the CLI manifest generation as code"""
    T = Transformer(manifest)
    console = Console()
    console.print(T.cli.code, overflow="fold", emoji=False, markup=False)
    out(f"# Rendered {T.cli.name} CLI v{T.cli.version} ~", fg="green")


@cli.command("run")  # type: ignore[arg-type]
@click.argument("manifest", type=click.File("rb"))
@click.argument("cli_args", type=str, nargs=-1)
def cliffy_run(manifest: TextIO, cli_args: tuple[str]) -> None:
    """Run CLI for a manifest"""
    T = Transformer(manifest)
    run_cli(T.cli.name, T.cli.code, cli_args)


@cli.command()  # type: ignore[arg-type]
@click.argument("cli_name", type=str, default="cliffy")
@click.option("--version", "-v", type=str, show_default=True, default="v1", help="Manifest version")
@click.option("--render", is_flag=True, show_default=True, default=False, help="Render template to terminal directly")
@click.option(
    "--raw",
    type=bool,
    is_flag=True,
    show_default=True,
    default=False,
    help="Raw template without boilerplate helpers and examples.",
)
def init(cli_name: str, version: str, render: bool, raw: bool) -> None:
    """Generate a CLI manifest template"""
    set_manifest_version(version)
    template = Manifest.get_raw_template(cli_name) if raw else Manifest.get_template(cli_name)

    if render:
        console = Console()
        console.print(template, overflow="fold", emoji=False, markup=False)
    else:
        try:
            write_to_file(f"{cli_name}.yaml", text=template)
        except Exception as e:
            exit_err(f"~ error writing to file: {e}")
        out(f"+ {cli_name}.yaml", fg="green")


@cli.command("list")  # type: ignore[arg-type]
def cliffy_list() -> None:
    """List all CLIs loaded"""
    cols = ["Name", "Version", "Age", "Manifest"]
    rows = [
        [metadata.cli_name, metadata.version, age_datetime(metadata.loaded), metadata.runner_path]
        for metadata in get_clis()
    ]
    print_rich_table(cols, rows, styles=["cyan", "magenta", "green", "blue"])


@cli.command()  # type: ignore[arg-type]
@click.argument("cli_names", type=str, nargs=-1)
def remove(cli_names: list[str]) -> None:
    """Remove a loaded CLI by name"""
    for cli_name in cli_names:
        if get_metadata_path(cli_name):
            remove_metadata(cli_name)
            Loader.unload_cli(cli_name)
            out(f"~ {cli_name} removed 💥", fg="green")
        else:
            out_err(f"~ {cli_name} not loaded")


@cli.command()  # type: ignore[arg-type]
def remove_all() -> None:
    """Remove all loaded CLIs"""
    for metadata in get_clis():
        remove_metadata(metadata.cli_name)
        Loader.unload_cli(metadata.cli_name)
        out(f"~ {metadata.cli_name} removed 💥", fg="green")


@cli.command()  # type: ignore[arg-type]
@click.argument("cli_or_manifests", type=ManifestOrCLI(), nargs=-1)
@click.option("--output-dir", "-o", type=click.Path(file_okay=False, dir_okay=True, writable=True), help="Output dir")
@click.option(
    "--python",
    "-p",
    type=str,
    help="Python interpreter to set as the zipapp shebang. This gets added after the #! ",
    default="/usr/bin/env python3",
    show_default=True,
)
def build(cli_or_manifests: list[Union[TextIOWrapper, str]], output_dir: str, python: str) -> None:
    """Build a CLI manifest or a loaded CLI into a zipapp"""
    for cli_or_manifest in cli_or_manifests:
        if isinstance(cli_or_manifest, TextIOWrapper):
            cli, result = build_cli_from_manifest(cli_or_manifest, output_dir=output_dir, interpreter=python)
            cli_name = cli.name
        else:
            cli_name = cast("str", cli_or_manifest)
            if not (metadata := get_metadata(cli_name)):
                out_err(f"~ {cli_name} not loaded")
                continue

            result = build_cli(
                cli_name,
                script_path=f"{CLIFFY_CLI_DIR}/{cli_name}.py",
                deps=metadata.requires,
                output_dir=output_dir,
                interpreter=python,
            )

        if result.exit_code != 0:
            out(result.stdout)
            out_err(f"~ {cli_name} build failed")
            continue

        out(f"+ {cli_name} built 📦", fg="green")


@cli.command()  # type: ignore[arg-type]
@click.argument("cli_name", type=str)
def info(cli_name: str):
    """Display CLI info"""
    metadata = get_metadata(cli_name) or exit_err(f"~ {cli_name} not loaded")
    out(f"{click.style('name:', fg='blue')} {metadata.cli_name}")
    out(f"{click.style('version:', fg='blue')} {metadata.version}")
    out(f"{click.style('requires:', fg='blue')} {metadata.requires}")
    out(f"{click.style('age:', fg='blue')} {age_datetime(metadata.loaded)} ({metadata.loaded.ctime()})")
    out(f"{click.style('manifest:', fg='blue')}\n{indent_block(metadata.manifest, spaces=2)}")


@cli.command()  # type: ignore[arg-type]
@click.argument("manifest", type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True))
@click.option(
    "--run-cli",
    type=str,
    default=False,
    help="If passed, runs CLI on each reload. Useful for syntax checks or testing command execution on each reload",
    is_flag=True,
)
@click.argument("run-cli-args", type=str, nargs=-1)
def dev(manifest: str, run_cli: bool, run_cli_args: tuple[str]) -> None:
    """Start hot-reloader for a manifest for active development.

    Examples:

    - cli dev examples/hello.yaml

    - cli dev examples/hello.yaml --run-cli  -- -h

    - cli dev examples/hello.yaml --run-cli hello
    """
    out(f"Watching {manifest} for changes\n", fg="magenta")
    CLIManifestReloader.watch(manifest, run_cli, run_cli_args)


ALIASES = {"add": load, "rm": remove, "rm-all": remove_all, "ls": cliffy_list, "reload": update}



import contextlib
import operator
import os
import platform
import subprocess
import sys
from datetime import datetime
from importlib.resources import files
from pathlib import Path
from tempfile import _TemporaryFileWrapper
from typing import Any, NoReturn, Optional

from click import secho
from packaging import version
from pydantic import BaseModel


CLIFFY_CLI_DIR = files("cliffy").joinpath("clis")
CLIFFY_METADATA_DIR = files("cliffy").joinpath("metadata")
PYTHON_BIN = (
    f"{os.path.join(sys.exec_prefix, 'Scripts')}"
    if platform.system() == "Windows"
    else f"{os.path.join(sys.exec_prefix, 'bin')}"
)
PYTHON_EXECUTABLE = sys.executable
OPERATOR_MAP = {
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    "<=": operator.le,
    "<": operator.lt,
    ">": operator.gt,
}
TEMP_FILES: list[_TemporaryFileWrapper] = []


class RequirementSpec(BaseModel):
    name: str
    operator: Optional[str]
    version: Optional[str]


def write_to_file(path: str, text: str, executable: bool = False) -> None:
    output_file = Path(path)
    output_file.parent.mkdir(exist_ok=True, parents=True)
    output_file.write_text(text)

    if executable:
        make_executable(path)


def make_executable(path: str) -> None:
    mode = os.stat(path).st_mode
    mode |= (mode & 0o444) >> 2
    os.chmod(path, mode)


def delete_temp_files() -> None:
    for file in TEMP_FILES:
        with contextlib.suppress(Exception):
            file.close()
            os.unlink(file.name)


def indent_block(block: str, spaces=4) -> str:
    return "\n".join([" " * spaces + line for line in block.splitlines()])


def wrap_as_comment(text: str, split_on: Optional[str] = None) -> str:
    if split_on:
        joiner = "\n# "
        return f"# {joiner.join(text.split(split_on))}"

    return f"# {text}"


def wrap_as_var(text: str) -> str:
    return "{{" + text + "}}"


def get_installed_package_versions() -> dict[str, str]:
    reqs = subprocess.check_output([sys.executable, "-m", "pip", "freeze"])
    installed_packages = {}
    for r in reqs.split():
        r_spec = r.decode().lower().split("==")
        if len(r_spec) > 1:
            installed_packages[r_spec[0]] = r_spec[1]
    return installed_packages


def parse_requirement(requirement: str) -> RequirementSpec:
    for op in OPERATOR_MAP:
        if op in requirement:
            parts = requirement.replace(" ", "").split(op)
            return RequirementSpec(name=parts[0], operator=op, version=parts[1])

    return RequirementSpec(name=requirement.strip(), operator=None, version=None)


def compare_versions(version1: str, version2: str, op: Optional[str] = "=") -> bool:
    v1 = version.parse(version1)
    v2 = version.parse(version2)

    return OPERATOR_MAP[op](v1, v2) if op else v1 == v2


def out(text: str, **echo_kwargs: Any) -> None:
    secho(text, **echo_kwargs)


def out_err(text: str) -> None:
    secho(f"{text} 💔", fg="red", err=True)


def exit_err(text: str) -> NoReturn:
    secho(f"{text} 💔", fg="red", err=True)
    raise SystemExit(1)


def age_datetime(date: datetime) -> str:
    delta = datetime.now() - date
    if delta.days > 0:
        return f"{delta.days}d"
    elif delta.seconds > 3600:
        return f"{delta.seconds // 3600}h"
    elif delta.seconds > 60:
        return f"{delta.seconds // 60}m"
    return f"{delta.seconds}s"



from io import TextIOWrapper
import os
import sys
from importlib import import_module
from pathlib import Path
from shutil import copy
from tempfile import NamedTemporaryFile, TemporaryDirectory
from types import ModuleType
from typing import Optional

from click.testing import CliRunner, Result
from shiv import cli as shiv_cli
from shiv import pip

from cliffy.helper import TEMP_FILES, delete_temp_files

from cliffy.transformer import Transformer

from cliffy.commander import CLI


def build_cli_from_manifest(
    manifestIO: TextIOWrapper, output_dir: Optional[str] = None, interpreter: str = "/usr/bin/env python3"
) -> tuple[CLI, Result]:
    T = Transformer(manifestIO, validate_requires=False)
    with NamedTemporaryFile(mode="w", prefix=f"{T.cli.name}_", suffix=".py", delete=False) as script:
        script.write(T.cli.code)
        script.flush()
        result = build_cli(
            T.cli.name, script_path=script.name, deps=T.cli.requires, output_dir=output_dir, interpreter=interpreter
        )
        TEMP_FILES.append(script)

    delete_temp_files()
    return T.cli, result


def build_cli(
    cli_name: str,
    script_path: str,
    deps: Optional[list[str]] = None,
    output_dir: Optional[str] = None,
    interpreter: str = "/usr/bin/env python3",
) -> Result:
    if deps is None:
        deps = []

    if output_dir and not Path(output_dir).exists():
        os.mkdir(output_dir)

    with TemporaryDirectory() as tdist:
        copy(script_path, os.path.join(tdist, f"{cli_name}.py"))
        pip_deps = ["typer"] + deps

        pip.install(["--target", tdist] + pip_deps)

        runner = CliRunner()
        output_file = os.path.join(output_dir, f"{cli_name}") if output_dir else cli_name
        return runner.invoke(
            shiv_cli.main,  # type: ignore
            ["--site-packages", tdist, "--compressed", "-e", f"{cli_name}.cli", "-o", output_file, "-p", interpreter],
        )


def import_module_from_path(filepath: str) -> ModuleType:
    module_path, module_filename = os.path.split(filepath)
    sys.path.append(module_path)
    return import_module(module_filename[:-3])


def run_cli(cli_name: str, script_code: str, args: tuple) -> None:
    with NamedTemporaryFile(mode="w", prefix=f"{cli_name}_", suffix=".py", delete=False) as runner_file:
        runner_file.write(script_code)
        runner_file.flush()
        module = import_module_from_path(runner_file.name)
        TEMP_FILES.append(runner_file)

    delete_temp_files()
    runner_argvs = [runner_file.name] + list(args)
    sys.argv = runner_argvs
    module.cli()
